require "defines"
require "util"

local builder = require "building.builder"
local collections = require "utility.collections"

--library classes for pathfinding
--local Grid = require ("jumper.grid")
local Pathfinder = require ("jumper.pathfinder")

local Map = require("utility.map")

local function OilWellNode(id, position)
    local self = {}
    self.id = id
    self.x = position.x
    self.y = position.y
    return self
end

local function absDist(a,b)
    trace(a==nil,"a")
    trace(b==nil,"b")
    return math.abs(a.x-b.x)+math.abs(a.y-b.y)
end

local function spotPosition(w)
    --{ {1, -2}, {2, -1}, {-1, 2}, {-2, 1} }
    if w.direction==nil then
        error("direction uninitialised: "..serpent.block(w))
    elseif w.direction == defines.direction.north then
        return {x = w.x+1 ,y = w.y-2}--x
    elseif w.direction == defines.direction.east then
        return {x = w.x+2 ,y = w.y-1}--y
    elseif w.direction == defines.direction.south then
        return {x = w.x-1 ,y = w.y+2}--x
    elseif w.direction == defines.direction.west then
        return {x = w.x-2 ,y = w.y+1}--y
    end
end

local function findSpot(map, well, lineEnd)
    local dx, dy = lineEnd.x-well.x, lineEnd.y-well.y
    if well.direction ~= nil then
        -- direction fixed, nothing changeable
        return spotPosition(well)
    elseif math.abs(dx)>math.abs(dy) then
        if dx < 0 then
            well.direction = defines.direction.west
        else
            well.direction = defines.direction.east
        end
    else
        if dy < 0 then
            well.direction = defines.direction.north
        else
            well.direction = defines.direction.south
        end
    end
    for i=0,2 do
        local spot = spotPosition(well)
        local tile = map:getTileState(spot.x, spot.y)
        if tile == map.tileStates.well or tile == map.tileStates.obstacle then
            well.direction = (well.direction + 2) % 8
        else
            break
        end
    end
    return spotPosition(well)
end
    
local function connectOilWells(player, spots, destination)
    local wells = {}
    local trees = collections.UnionFind()
    local nearest, nearestDist
    
    local map = Map.new()
    
    for _, e in pairs(spots) do
        wells[#wells+1] = OilWellNode(#wells+1, e.position)
        trees.makeSet(#wells)
        if nearest == nil or absDist(destination, wells[#wells]) < nearestDist then
            nearestDist = absDist(destination, wells[#wells])
            nearest = wells[#wells]
        end
        map:addWell(wells[#wells])
    end
    
    map:addPipe(destination,nearest.id)
    
    --local gridMap, offsetX,offsetY = map:createGridMap()
    --local grid = Grid(gridMap)
    -- Creates a pathfinder object using Jump Point Search
    local myFinder = Pathfinder(map, 'JPS', function(tile)
        return tile==map.tileStates.empty or tile==map.tileStates.pipe
    end)
    myFinder:setMode('ORTHOGONAL')
    
    local function connect(start, ziel, startID, endID)
        startID = trees.findSet(startID)
        -- Calculates the path, and its length
        local path, length = myFinder:getPath(start.x, start.y, ziel.x, ziel.y)
        if path then
            --print(('Path found! Length: %.2f'):format(length))
            for node, count in path:iter() do
                --print(('Step: %d - x: %d - y: %d'):format(count, node.x, node.y))
                builder.buildEntity({name = "pipe", position = node, force = player.force}, true)
                for i=-1,1 do
                    for j=-1,1 do
                        if (map:getTileState(node.x+i,node.y+j) == map.tileStates.pipe) then
                           trees.checkedUnion(startID,map:getTile(node.x+i,node.y+j).id) 
                        end
                    end
                end
                map:addPipe(node,startID)
            end
        end
        
        trees.checkedUnion(startID, endID) -- O(n)
    end
    
    local function connectWW(e)
        if e.a.direction == nil and e.a.direction == nil then
            local tmp = findSpot(map, e.a, e.b) -- use center of well as orientation
            return connect(tmp, findSpot(map, e.b, tmp), trees.findSet(e.a.id), trees.findSet(e.b.id))
        elseif e.b.direction == nil then
            local tmp = spotPosition(e.a)
            return connect(tmp, findSpot(map, e.b, tmp), trees.findSet(e.a.id), trees.findSet(e.b.id))
        elseif e.a.direction == nil then
            local tmp = spotPosition(e.b)
            return connect(findSpot(map, e.a, tmp), tmp, trees.findSet(e.a.id), trees.findSet(e.b.id))
        else
            return connect(spotPosition(e.a), spotPosition(e.b), trees.findSet(e.a.id), trees.findSet(e.b.id))
        end
    end
    
    local conHeap = collections.Heap(0)
    local meta = {
        __eq = function (lhs, rhs)
            return lhs.dist == rhs.dist
          end,
        __lt = function (lhs, rhs)
            return lhs.dist < rhs.dist
        end
    }
    for i=1,#wells do
        for j=i+1, #wells do
            conHeap[#conHeap + 1] = setmetatable({a=wells[i],b=wells[j],dist=absDist(wells[i],wells[j])}, meta)
        end
    end
    conHeap.size = #conHeap
    conHeap.buildMaxHeap() -- O(n) needed before sorting
    conHeap.sort() -- O(n*log n)
    
    --kruskal
    connect(destination, findSpot(nearest,destination), nearest.id, nearest.id)
    local i = 1
    while trees.maxSize < #wells do -- and i <= #conHeap
        local e = conHeap[i]
        if trees.findSet(e.a.id) ~= trees.findSet(e.b.id) then
            connectWW(e)
            coroutine.yield()
        end
        i = i+1
    end
    
    for i, well in ipairs(wells) do
        builder.buildEntity({name = "pumpjack", position = well, 
                force = player.force, direction = well.direction}, true)
        if i%10 == 0 then
            coroutine.yield()
        end
    end
    map:save()
end

return {
    name = "Oil Outpost",
    options = {
        {name = "Oilfield", type = "Polygon" },
        {name = "Refinery Start", type = "Position" },
        {name = "Build Refineries", type = "checkbox", selected = true},
        {name = "Refinery Count", type = "textfield", value = "5" },
        {name = "BuildDir", type = "Option", values = {"North","East","South","West"}, selected = 1},
    },
    build = function (player, options)
        local field = options[1]
        local refineryStart = options[2]
        local buildRefineries = options[3]
        local count = tonumber(options[4])
        local dir = options[5]
        if not (field and refineryStart and (not buildRefineries or (count and count > 0))) then
            player.print("Invalid Parameter for Oil generator: "..serpent.block(options))
            return false -- invalid/missing arguments
        end
        local spots = field:findEntities("crude-oil")
        connectOilWells(player, spots, refineryStart)
        if buildRefineries then
            builder.buildStructure("refinery", refineryStart, 2 * (dir - 1), count)
        end
    end
}