local collections = {}

function collections.Heap(size)
    local self = {}
    if type(size)=="table" then -- copy constructor
        self.size = size.size
        for k,v in ipairs(size) do
            self[k] = v
        end
    else -- init constructor
        self.size = size
    end

    self.parent = function(i)
        return math.floor (i/2)
    end
    self.left = function(i)
        return 2*i
    end
    self.right = function(i)
        return 2*i+1
    end
    
    local function swap(a,b)
        local tmp = self[a]
        self[a] = self[b]
        self[b] = tmp
    end
    
    ---
    -- Updates max heap if element i changed
    -- O( log(n) )
    self.maxHeapify = function(i)
        local l, r = self.left(i), self.right(i)
        local largest = i
        if l <= self.size and self[l] > self[i] then
            largest = l
        end
        if r<= self.size and self[r] > self[largest] then
            largest = r
        end
        if largest ~= i then
            swap(i,largest)
            self.maxHeapify(largest)
        end
    end
    
    ---
    -- Orders unsorted array to max heap
    -- O(n)
    self.buildMaxHeap = function()
        for i = math.floor(self.size/2), 1, -1 do
            self.maxHeapify(i)
        end
    end
    
    ---
    -- Reorders heap to a sorted Array
    -- O(n*log(n))
    self.sort = function()
        local originalSize = self.size
        for i=self.size,2,-1 do
            swap(i,1)
            self.size = self.size-1
            self.maxHeapify(1)            
        end
        self.size = originalSize
    end
    
    return self
end

function collections.UnionFind()
    local self = {} -- methods
    local nodes = {} --map x <-> {rep = representant, size, next}
    self.maxSize = 1
    ---
    -- adds a number to the sets
    --  O(1)
    self.makeSet = function(x)
        nodes[x] = {rep = x, size = 1, next=nil}
    end
    ---
    -- finds the first element (representant) of the set x is in
    --  O(1)
    self.findSet = function(x)
        if type(x) ~= "number" then
            trace()
            game.player.print(serpent.block(x))
        end
        return nodes[x].rep
    end
    
    ---
    -- unions the two sets x and y are in (must be disjunct)
    --  O (n)
    self.union = function(x,y)
        local a,b = self.findSet(x), self.findSet(y)
        if a < b then do local tmp = a; a=b; b=tmp; end end
        
        nodes[a].size = nodes[a].size + nodes[b].size
        if(nodes[a].size > self.maxSize) then
            self.maxSize = nodes[a].size
        end
        
        local c = nodes[a].next
        nodes[a].next = b
        
        while nodes[b].next ~= nil do
            nodes[b].rep = a
            b = nodes[b].next
        end
        nodes[b].rep = a
        nodes[b].next = c
        return a
    end
    
    self.checkedUnion = function(x,y)
        if self.findSet(x) ~= self.findSet(y) then
            self.union(x,y)
        end
    end
    
    self.size = function(x)
        return nodes[self.findSet(x)].size
    end
    return self
end

return collections